"""
Useful additional string functions.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, cast

if TYPE_CHECKING:
    from typing import Any, Union


def remove_non_ascii(s: str) -> str:
    """
    Remove non-ascii characters in a file. Needed when support for non-ASCII
    is not available.

    Args:
        s (str): Input string

    Returns:
        String with all non-ascii characters removed.
    """
    return "".join(i for i in s if ord(i) < 128)


def is_string(s: Any) -> bool:
    """True if s behaves like a string (duck typing test)."""
    try:
        s + " "
        return True

    except TypeError:
        return False


def list_strings(arg: Union[str, Iterable[str]]) -> list[str]:
    """
    Always return a list of strings, given a string or list of strings as
    input.

    Examples:
        >>> list_strings('A single string')
        ['A single string']

        >>> list_strings(['A single string in a list'])
        ['A single string in a list']

        >>> list_strings(['A','list','of','strings'])
        ['A', 'list', 'of', 'strings']

        >>> list_strings(('A','list','of','strings'))
        ['A', 'list', 'of', 'strings']

        >>> list_strings({"a": 1, "b": 2}.keys())
        ['a', 'b']
    """
    if is_string(arg):
        return [cast(str, arg)]

    return [cast(str, s) for s in arg]


def marquee(text: str = "", width: int = 78, mark: str = "*") -> str:
    """
    Return the input string centered in a 'marquee'.

    Args:
        text (str): Input string
        width (int): Width of final output string.
        mark (str): Character used to fill string.

    Examples:
        >>> marquee('A test', width=40)
        '**************** A test ****************'

        >>> marquee('A test', width=40, mark='-')
        '---------------- A test ----------------'

        marquee('A test',40, ' ')
        '                 A test                 '
    """
    if not text:
        return (mark * width)[:width]

    nmark = (width - len(text) - 2) // len(mark) // 2
    nmark = max(nmark, 0)

    marks = mark * nmark
    return f"{marks} {text} {marks}"


def boxed(msg: str, ch: str = "=", pad: int = 5) -> str:
    """
    Returns a string in a box

    Args:
        msg: Input string.
        ch: Character used to form the box.
        pad: Number of characters ch added before and after msg.

    Examples:
        >>> print(boxed("hello", ch="*", pad=2))
        ***********
        ** hello **
        ***********
    """
    if pad > 0:
        msg = pad * ch + " " + msg.strip() + " " + pad * ch

    return "\n".join(
        [
            len(msg) * ch,
            msg,
            len(msg) * ch,
        ]
    )


def make_banner(s: str, width: int = 78, mark: str = "*") -> str:
    """
    Args:
        s: String
        width: Width of banner. Defaults to 78.
        mark: The mark used to create the banner.

    Returns:
        Banner string.
    """
    banner = marquee(s, width=width, mark=mark)
    return "\n" + len(banner) * mark + "\n" + banner + "\n" + len(banner) * mark


def indent(lines: str, amount: int, ch: str = " ") -> str:
    """
    Indent the lines in a string by padding each one with proper number of pad
    characters
    """
    padding = amount * ch
    return padding + ("\n" + padding).join(lines.split("\n"))



"""
Path based methods, e.g., which, zpath, etc.
"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from monty.fnmatch import WildCard
from monty.string import list_strings

if TYPE_CHECKING:
    from typing import Callable, Literal, Optional, Union


def zpath(filename: str) -> str:
    """
    Returns an existing (zipped or unzipped) file path given the unzipped
    version. If no path exists, returns the filename unmodified.

    Args:
        filename: filename without zip extension

    Returns:
        filename with a zip extension (unless an unzipped version
        exists). If filename is not found, the same filename is returned
        unchanged.
    """
    exts = ("", ".gz", ".GZ", ".bz2", ".BZ2", ".z", ".Z")
    for ext in exts:
        filename = filename.removesuffix(ext)

    for ext in exts:
        zfilename = f"{filename}{ext}"
        if os.path.exists(zfilename):
            return zfilename
    return filename


def find_exts(
    top: str,
    exts: Union[str, list[str]],
    exclude_dirs: Optional[str] = None,
    include_dirs: Optional[str] = None,
    match_mode: Literal["basename", "abspath"] = "basename",
) -> list[str]:
    """
    Find all files with the extension listed in `exts` that are located within
    the directory tree rooted at `top` (including top itself, but excluding
    '.' and '..')

    Args:
        top (str): Root directory
        exts (str or list of strings): List of extensions.
        exclude_dirs (str): Wildcards used to exclude particular directories.
            Can be concatenated via `|`
        include_dirs (str): Wildcards used to select particular directories.
            `include_dirs` and `exclude_dirs` are mutually exclusive
        match_mode (str): "basename" if  match should be done on the basename.
            "abspath" for absolute path.

    Returns:
        list[str]: Absolute paths of the files.

    Examples:
        # Find all pdf and ps files starting from the current directory.
        find_exts(".", ("pdf", "ps"))

        # Find all pdf files, exclude hidden directories and dirs whose name
        # starts with `_`
        find_exts(".", "pdf", exclude_dirs="_*|.*")

        # Find all ps files, in the directories whose basename starts with
        # output.
        find_exts(".", "ps", include_dirs="output*"))
    """
    exts = list_strings(exts)

    # Handle file!
    if os.path.isfile(top):
        return [os.path.abspath(top)] if any(top.endswith(ext) for ext in exts) else []

    # Build shell-style wildcards.
    if exclude_dirs is not None:
        _exclude_dirs: WildCard = WildCard(exclude_dirs)

    if include_dirs is not None:
        _include_dirs: WildCard = WildCard(include_dirs)

    mangle_functions: dict[str, Callable[..., str]] = {
        "basename": os.path.basename,
        "abspath": os.path.abspath,
    }
    mangle: Callable[..., str] = mangle_functions[match_mode]

    # Assume directory
    paths = []
    for dirpath, _dirnames, filenames in os.walk(top):
        dirpath = os.path.abspath(dirpath)

        if exclude_dirs and _exclude_dirs.match(mangle(dirpath)):
            continue
        if include_dirs and not _include_dirs.match(mangle(dirpath)):
            continue

        for filename in filenames:
            if any(filename.endswith(ext) for ext in exts):
                paths.append(os.path.join(dirpath, filename))

    return paths



"""
Multiprocessing utilities.
"""

from __future__ import annotations

from multiprocessing import Pool
from typing import Callable, Iterable

try:
    from tqdm.autonotebook import tqdm
except ImportError:
    tqdm = None


def imap_tqdm(nprocs: int, func: Callable, iterable: Iterable, *args, **kwargs) -> list:
    """
    A wrapper around Pool.imap. Creates a Pool with nprocs and then runs a f
    unction over an iterable with progress bar.

    Args:
        nprocs: Number of processes
        func: Callable
        iterable: Iterable of arguments.
        args: Passthrough to Pool.imap
        kwargs: Passthrough to Pool.imap

    Returns:
        Results of Pool.imap.
    """
    if tqdm is None:
        raise ImportError("tqdm must be installed for this function.")
    data = []
    with Pool(nprocs) as pool:
        try:
            n = len(iterable)  # type: ignore
        except TypeError:
            n = None  # type: ignore
        with tqdm(total=n) as prog_bar:
            for d in pool.imap(func, iterable, *args, **kwargs):
                prog_bar.update()
                data.append(d)
    return data



"""
JSON serialization and deserialization utilities.
"""

from __future__ import annotations

import datetime
import json
import os
import pathlib
import pickle
import traceback
import types
from collections import OrderedDict, defaultdict
from enum import Enum
from hashlib import sha1
from importlib import import_module
from inspect import getfullargspec
from pathlib import Path
from typing import Any, Dict
from uuid import UUID, uuid4

try:
    import numpy as np
except ImportError:
    np = None  # type: ignore

try:
    import pydantic
except ImportError:
    pydantic = None  # type: ignore

try:
    from pydantic_core import core_schema
except ImportError:
    core_schema = None  # type: ignore

try:
    import bson
except ImportError:
    bson = None  # type: ignore

try:
    from ruamel.yaml import YAML
except ImportError:
    YAML = None  # type: ignore

try:
    import orjson
except ImportError:
    orjson = None  # type: ignore

try:
    import dataclasses
except ImportError:
    dataclasses = None  # type: ignore

try:
    import torch
except ImportError:
    torch = None  # type: ignore

__version__ = "3.0.0"


def _load_redirect(redirect_file):
    try:
        with open(redirect_file) as f:
            yaml = YAML()
            d = yaml.load(f)
    except OSError:
        # If we can't find the file
        # Just use an empty redirect dict
        return {}

    # Convert the full paths to module/class
    redirect_dict = defaultdict(dict)
    for old_path, new_path in d.items():
        old_class = old_path.split(".")[-1]
        old_module = ".".join(old_path.split(".")[:-1])

        new_class = new_path.split(".")[-1]
        new_module = ".".join(new_path.split(".")[:-1])

        redirect_dict[old_module][old_class] = {
            "@module": new_module,
            "@class": new_class,
        }

    return dict(redirect_dict)


def _check_type(obj, type_str) -> bool:
    """Alternative to isinstance that avoids imports.

    Checks whether obj is an instance of the type defined by type_str. This
    removes the need to explicitly import type_str. Handles subclasses like
    isinstance does. E.g.::
        class A:
            pass


        class B(A):
            pass


        a, b = A(), B()
        assert isinstance(a, A)
        assert isinstance(b, B)
        assert isinstance(b, A)
        assert not isinstance(a, B)

    type_str: str | tuple[str]

    Note for future developers: the type_str is not always obvious for an
    object. For example, pandas.DataFrame is actually pandas.core.frame.DataFrame.
    To find out the type_str for an object, run type(obj).mro(). This will
    list all the types that an object can resolve to in order of generality
    (all objects have the builtins.object as the last one).
    """
    type_str = type_str if isinstance(type_str, tuple) else (type_str,)
    # I believe this try-except is only necessary for callable types
    try:
        mro = type(obj).mro()
    except TypeError:
        return False
    return any(o.__module__ + "." + o.__name__ == ts for o in mro for ts in type_str)


class MSONable:
    """
    This is a mix-in base class specifying an API for msonable objects. MSON
    is Monty JSON. Essentially, MSONable objects must implement an as_dict
    method, which must return a json serializable dict and must also support
    no arguments (though optional arguments to finetune the output is ok),
    and a from_dict class method that regenerates the object from the dict
    generated by the as_dict method. The as_dict method should contain the
    "@module" and "@class" keys which will allow the MontyEncoder to
    dynamically deserialize the class. E.g.::

        d["@module"] = self.__class__.__module__
        d["@class"] = self.__class__.__name__

    A default implementation is provided in MSONable, which automatically
    determines if the class already contains self.argname or self._argname
    attributes for every arg. If so, these will be used for serialization in
    the dict format. Similarly, the default from_dict will deserialization
    classes of such form. An example is given below::

        class MSONClass(MSONable):

        def __init__(self, a, b, c, d=1, **kwargs):
            self.a = a
            self.b = b
            self._c = c
            self._d = d
            self.kwargs = kwargs

    For such classes, you merely need to inherit from MSONable and you do not
    need to implement your own as_dict or from_dict protocol.

    New to Monty V2.0.6....
    Classes can be redirected to moved implementations by putting in the old
    fully qualified path and new fully qualified path into .monty.yaml in the
    home folder

    Example:
    old_module.old_class: new_module.new_class
    """

    REDIRECT = _load_redirect(os.path.join(os.path.expanduser("~"), ".monty.yaml"))

    def as_dict(self) -> dict:
        """
        A JSON serializable dict representation of an object.
        """
        d = {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,
        }

        try:
            parent_module = self.__class__.__module__.split(".", maxsplit=1)[0]
            module_version = import_module(parent_module).__version__  # type: ignore
            d["@version"] = str(module_version)
        except (AttributeError, ImportError):
            d["@version"] = None  # type: ignore

        spec = getfullargspec(self.__class__.__init__)

        def recursive_as_dict(obj):
            if isinstance(obj, (list, tuple)):
                return [recursive_as_dict(it) for it in obj]
            if isinstance(obj, dict):
                return {kk: recursive_as_dict(vv) for kk, vv in obj.items()}
            if hasattr(obj, "as_dict"):
                return obj.as_dict()
            if dataclasses is not None and dataclasses.is_dataclass(obj):
                d = dataclasses.asdict(obj)
                d.update(
                    {
                        "@module": obj.__class__.__module__,
                        "@class": obj.__class__.__name__,
                    }
                )
                return d
            return obj

        for c in spec.args + spec.kwonlyargs:
            if c != "self":
                try:
                    a = getattr(self, c)
                except AttributeError:
                    try:
                        a = getattr(self, "_" + c)
                    except AttributeError:
                        raise NotImplementedError(
                            "Unable to automatically determine as_dict "
                            "format from class. MSONAble requires all "
                            "args to be present as either self.argname or "
                            "self._argname, and kwargs to be present under "
                            "a self.kwargs variable to automatically "
                            "determine the dict format. Alternatively, "
                            "you can implement both as_dict and from_dict."
                        )
                d[c] = recursive_as_dict(a)
        if hasattr(self, "kwargs"):
            # type: ignore
            d.update(**self.kwargs)  # pylint: disable=E1101
        if spec.varargs is not None and getattr(self, spec.varargs, None) is not None:
            d.update({spec.varargs: getattr(self, spec.varargs)})
        if hasattr(self, "_kwargs"):
            d.update(**self._kwargs)  # pylint: disable=E1101
        if isinstance(self, Enum):
            d.update({"value": self.value})  # pylint: disable=E1101
        return d

    @classmethod
    def from_dict(cls, d):
        """
        :param d: Dict representation.
        :return: MSONable class.
        """
        decoded = {
            k: MontyDecoder().process_decoded(v)
            for k, v in d.items()
            if not k.startswith("@")
        }
        return cls(**decoded)

    def to_json(self) -> str:
        """
        Returns a json string representation of the MSONable object.
        """
        return json.dumps(self, cls=MontyEncoder)

    def unsafe_hash(self):
        """
        Returns an hash of the current object. This uses a generic but low
        performance method of converting the object to a dictionary, flattening
        any nested keys, and then performing a hash on the resulting object
        """

        def flatten(obj, separator="."):
            # Flattens a dictionary

            flat_dict = {}
            for key, value in obj.items():
                if isinstance(value, dict):
                    flat_dict.update(
                        {
                            separator.join([key, _key]): _value
                            for _key, _value in flatten(value).items()
                        }
                    )
                elif isinstance(value, list):
                    list_dict = {
                        f"{key}{separator}{num}": item for num, item in enumerate(value)
                    }
                    flat_dict.update(flatten(list_dict))
                else:
                    flat_dict[key] = value

            return flat_dict

        ordered_keys = sorted(
            flatten(jsanitize(self.as_dict())).items(), key=lambda x: x[0]
        )
        ordered_keys = [item for item in ordered_keys if "@" not in item[0]]
        return sha1(json.dumps(OrderedDict(ordered_keys)).encode("utf-8"))

    @classmethod
    def _validate_monty(cls, __input_value):
        """
        pydantic Validator for MSONable pattern
        """
        if isinstance(__input_value, cls):
            return __input_value
        if isinstance(__input_value, dict):
            # Do not allow generic exceptions to be raised during deserialization
            # since pydantic may handle them incorrectly.
            try:
                new_obj = MontyDecoder().process_decoded(__input_value)
                if isinstance(new_obj, cls):
                    return new_obj
                return cls(**__input_value)
            except Exception:
                raise ValueError(
                    f"Error while deserializing {cls.__name__} "
                    f"object: {traceback.format_exc()}"
                )

        raise ValueError(
            f"Must provide {cls.__name__}, the as_dict form, or the proper"
        )

    @classmethod
    def validate_monty_v1(cls, __input_value):
        """
        Pydantic validator with correct signature for pydantic v1.x
        """
        return cls._validate_monty(__input_value)

    @classmethod
    def validate_monty_v2(cls, __input_value, _):
        """
        Pydantic validator with correct signature for pydantic v2.x
        """
        return cls._validate_monty(__input_value)

    @classmethod
    def __get_validators__(cls):
        """Return validators for use in pydantic"""
        yield cls.validate_monty_v1

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type, handler):
        """
        pydantic v2 core schema definition
        """
        if core_schema is None:
            raise RuntimeError("Pydantic >= 2.0 is required for validation")

        s = core_schema.with_info_plain_validator_function(cls.validate_monty_v2)

        return core_schema.json_or_python_schema(json_schema=s, python_schema=s)

    @classmethod
    def _generic_json_schema(cls):
        return {
            "type": "object",
            "properties": {
                "@class": {"enum": [cls.__name__], "type": "string"},
                "@module": {"enum": [cls.__module__], "type": "string"},
                "@version": {"type": "string"},
            },
            "required": ["@class", "@module"],
        }

    @classmethod
    def __get_pydantic_json_schema__(cls, core_schema, handler):
        """JSON schema for MSONable pattern"""
        return cls._generic_json_schema()

    @classmethod
    def __modify_schema__(cls, field_schema):
        """JSON schema for MSONable pattern"""
        custom_schema = cls._generic_json_schema()
        field_schema.update(custom_schema)

    def _get_partial_json(self, json_kwargs, pickle_kwargs):
        """Used with the save method. Gets the json representation of a class
        with the unserializable components sustituted for hash references."""

        if pickle_kwargs is None:
            pickle_kwargs = {}
        if json_kwargs is None:
            json_kwargs = {}
        encoder = MontyEncoder(allow_unserializable_objects=True, **json_kwargs)
        encoded = encoder.encode(self)
        return encoder, encoded, json_kwargs, pickle_kwargs

    def get_partial_json(self, json_kwargs=None, pickle_kwargs=None):
        """
        Parameters
        ----------
        json_kwargs : dict
            Keyword arguments to pass to the serializer.
        pickle_kwargs : dict
            Keyword arguments to pass to pickle.dump.

        Returns
        -------
        str, dict
            The json encoding of the class and the name-object map if one is
            required, otherwise None.
        """

        encoder, encoded, json_kwargs, pickle_kwargs = self._get_partial_json(
            json_kwargs, pickle_kwargs
        )
        name_object_map = encoder._name_object_map
        if len(name_object_map) == 0:
            name_object_map = None
        return encoded, name_object_map, json_kwargs, pickle_kwargs

    def save(
        self,
        json_path,
        mkdir=True,
        json_kwargs=None,
        pickle_kwargs=None,
        strict=True,
    ):
        """Utility that uses the standard tools of MSONable to convert the
        class to json format, but also save it to disk. In addition, this
        method intelligently uses pickle to individually pickle class objects
        that are not serializable, saving them separately. This maximizes the
        readability of the saved class information while allowing _any_
        class to be at least partially serializable to disk.

        For a fully MSONable class, only a class.json file will be saved to
        the location {save_dir}/class.json. For a partially MSONable class,
        additional information will be saved to the save directory at
        {save_dir}. This includes a pickled object for each attribute that
        e serialized.

        Parameters
        ----------
        file_path : os.PathLike
            The file to which to save the json object. A pickled object of
            the same name but different extension might also be saved if the
            class is not entirely MSONable.
        mkdir : bool
            If True, makes the provided directory, including all parent
            directories.
        json_kwargs : dict
            Keyword arguments to pass to the serializer.
        pickle_kwargs : dict
            Keyword arguments to pass to pickle.dump.
        strict : bool
            If True, will not allow you to overwrite existing files.
        """

        json_path = Path(json_path)
        save_dir = json_path.parent

        encoded, name_object_map, json_kwargs, pickle_kwargs = self.get_partial_json(
            json_kwargs, pickle_kwargs
        )

        if mkdir:
            save_dir.mkdir(exist_ok=True, parents=True)

        # Define the pickle path
        pickle_path = save_dir / f"{json_path.stem}.pkl"

        # Check if the files exist and the strict parameter is True
        if strict and json_path.exists():
            raise FileExistsError(f"strict is true and file {json_path} exists")
        if strict and pickle_path.exists():
            raise FileExistsError(f"strict is true and file {pickle_path} exists")

        # Save the json file
        with open(json_path, "w") as outfile:
            outfile.write(encoded)

        # Save the pickle file if we have anything to save from the name_object_map
        if name_object_map is not None:
            with open(pickle_path, "wb") as f:
                pickle.dump(name_object_map, f, **pickle_kwargs)

    @classmethod
    def load(cls, file_path):
        """Loads a class from a provided json file.

        Parameters
        ----------
        file_path : os.PathLike
            The json file to load from.

        Returns
        -------
        MSONable
            An instance of the class being reloaded.
        """

        d = _d_from_path(file_path)
        return cls.from_dict(d)


def load(path):
    """Loads a json file that was saved using MSONable.save.

    Parameters
    ----------
    path : os.PathLike
        Path to the json file to load.

    Returns
    -------
    MSONable
    """

    d = _d_from_path(path)
    module = d["@module"]
    klass = d["@class"]
    module = import_module(module)
    klass = getattr(module, klass)
    return klass.from_dict(d)


def _d_from_path(file_path):
    json_path = Path(file_path)
    save_dir = json_path.parent
    pickle_path = save_dir / f"{json_path.stem}.pkl"

    with open(json_path, "r") as infile:
        d = json.loads(infile.read())

    if pickle_path.exists():
        name_object_map = pickle.load(open(pickle_path, "rb"))
        d = _recursive_name_object_map_replacement(d, name_object_map)
    return d


def _recursive_name_object_map_replacement(d, name_object_map):
    if isinstance(d, dict):
        if "@object_reference" in d:
            name = d["@object_reference"]
            return name_object_map.pop(name)
        return {
            k: _recursive_name_object_map_replacement(v, name_object_map)
            for k, v in d.items()
        }
    elif isinstance(d, list):
        return [_recursive_name_object_map_replacement(x, name_object_map) for x in d]
    return d


class MontyEncoder(json.JSONEncoder):
    """
    A Json Encoder which supports the MSONable API, plus adds support for
    numpy arrays, datetime objects, bson ObjectIds (requires bson).
    Usage::
        # Add it as a *cls* keyword when using json.dump
        json.dumps(object, cls=MontyEncoder)
    """

    def __init__(self, *args, allow_unserializable_objects=False, **kwargs):
        super().__init__(*args, **kwargs)
        self._allow_unserializable_objects = allow_unserializable_objects
        self._name_object_map: Dict[str, Any] = {}
        self._index = 0

    def _update_name_object_map(self, o):
        name = f"{self._index:012}-{str(uuid4())}"
        self._index += 1
        self._name_object_map[name] = o
        return {"@object_reference": name}

    def default(self, o) -> dict:  # pylint: disable=E0202
        """
        Overriding default method for JSON encoding. This method does two
        things: (a) If an object has a to_dict property, return the to_dict
        output. (b) If the @module and @class keys are not in the to_dict,
        add them to the output automatically. If the object has no to_dict
        property, the default Python json encoder default method is called.
        Args:
            o: Python object.
        Return:
            Python dict representation.
        """
        if isinstance(o, datetime.datetime):
            return {
                "@module": "datetime",
                "@class": "datetime",
                "string": str(o),
            }
        if isinstance(o, UUID):
            return {"@module": "uuid", "@class": "UUID", "string": str(o)}
        if isinstance(o, Path):
            return {"@module": "pathlib", "@class": "Path", "string": str(o)}

        if torch is not None and isinstance(o, torch.Tensor):
            # Support for Pytorch Tensors.
            d = {
                "@module": "torch",
                "@class": "Tensor",
                "dtype": o.type(),
            }
            if "Complex" in o.type():
                d["data"] = [o.real.tolist(), o.imag.tolist()]  # type: ignore
            else:
                d["data"] = o.numpy().tolist()
            return d

        if np is not None:
            if isinstance(o, np.ndarray):
                if str(o.dtype).startswith("complex"):
                    return {
                        "@module": "numpy",
                        "@class": "array",
                        "dtype": str(o.dtype),
                        "data": [o.real.tolist(), o.imag.tolist()],
                    }
                return {
                    "@module": "numpy",
                    "@class": "array",
                    "dtype": str(o.dtype),
                    "data": o.tolist(),
                }
            if isinstance(o, np.generic):
                return o.item()

        if _check_type(o, "pandas.core.frame.DataFrame"):
            return {
                "@module": "pandas",
                "@class": "DataFrame",
                "data": o.to_json(default_handler=MontyEncoder().encode),
            }
        if _check_type(o, "pandas.core.series.Series"):
            return {
                "@module": "pandas",
                "@class": "Series",
                "data": o.to_json(default_handler=MontyEncoder().encode),
            }

        if bson is not None and isinstance(o, bson.objectid.ObjectId):
            return {
                "@module": "bson.objectid",
                "@class": "ObjectId",
                "oid": str(o),
            }

        if callable(o) and not isinstance(o, MSONable):
            try:
                return _serialize_callable(o)
            except AttributeError as e:
                # Some callables may not have instance __name__
                if self._allow_unserializable_objects:
                    return self._update_name_object_map(o)
                raise AttributeError(e)

        try:
            if pydantic is not None and isinstance(o, pydantic.BaseModel):
                d = o.model_dump()
            elif (
                dataclasses is not None
                and (not issubclass(o.__class__, MSONable))
                and dataclasses.is_dataclass(o)
            ):
                # This handles dataclasses that are not subclasses of MSONAble.
                d = dataclasses.asdict(o)
            elif hasattr(o, "as_dict"):
                d = o.as_dict()
            elif isinstance(o, Enum):
                d = {"value": o.value}
            elif self._allow_unserializable_objects:
                # Last resort logic. We keep track of some name of the object
                # as a reference, and instead of the object, store that
                # name, which of course is json-serializable
                d = self._update_name_object_map(o)
            else:
                raise TypeError(
                    f"Object of type {o.__class__.__name__} is not JSON serializable"
                )

            if "@module" not in d:
                d["@module"] = str(o.__class__.__module__)
            if "@class" not in d:
                d["@class"] = str(o.__class__.__name__)
            if "@version" not in d:
                try:
                    parent_module = o.__class__.__module__.split(".")[0]
                    module_version = import_module(parent_module).__version__  # type: ignore
                    d["@version"] = str(module_version)
                except (AttributeError, ImportError):
                    d["@version"] = None  # type: ignore
            return d
        except AttributeError:
            return json.JSONEncoder.default(self, o)


class MontyDecoder(json.JSONDecoder):
    """
    A Json Decoder which supports the MSONable API. By default, the
    decoder attempts to find a module and name associated with a dict. If
    found, the decoder will generate a Pymatgen as a priority.  If that fails,
    the original decoded dictionary from the string is returned. Note that
    nested lists and dicts containing pymatgen object will be decoded correctly
    as well.

    Usage:

        # Add it as a *cls* keyword when using json.load
        json.loads(json_string, cls=MontyDecoder)
    """

    def process_decoded(self, d):
        """
        Recursive method to support decoding dicts and lists containing
        pymatgen objects.
        """
        if isinstance(d, dict):
            if "@module" in d and "@class" in d:
                modname = d["@module"]
                classname = d["@class"]
                if cls_redirect := MSONable.REDIRECT.get(modname, {}).get(classname):
                    classname = cls_redirect["@class"]
                    modname = cls_redirect["@module"]

            elif "@module" in d and "@callable" in d:
                modname = d["@module"]
                objname = d["@callable"]
                classname = None
                if d.get("@bound", None) is not None:
                    # if the function is bound to an instance or class, first
                    # deserialize the bound object and then remove the object name
                    # from the function name.
                    obj = self.process_decoded(d["@bound"])
                    objname = objname.split(".")[1:]
                else:
                    # if the function is not bound to an object, import the
                    # function from the module name
                    obj = __import__(modname, globals(), locals(), [objname], 0)
                    objname = objname.split(".")
                try:
                    # the function could be nested. e.g., MyClass.NestedClass.function
                    # so iteratively access the nesting
                    for attr in objname:
                        obj = getattr(obj, attr)

                    return obj

                except AttributeError:
                    pass
            else:
                modname = None
                classname = None

            if classname:
                if modname and modname not in {
                    "bson.objectid",
                    "numpy",
                    "pandas",
                    "torch",
                }:
                    if modname == "datetime" and classname == "datetime":
                        try:
                            # Remove timezone info in the form of "+xx:00"
                            dt = datetime.datetime.strptime(
                                d["string"].split("+")[0], "%Y-%m-%d %H:%M:%S.%f"
                            )
                        except ValueError:
                            dt = datetime.datetime.strptime(
                                d["string"].split("+")[0], "%Y-%m-%d %H:%M:%S"
                            )
                        return dt

                    elif modname == "uuid" and classname == "UUID":
                        return UUID(d["string"])

                    elif modname == "pathlib" and classname == "Path":
                        return Path(d["string"])

                    mod = __import__(modname, globals(), locals(), [classname], 0)
                    if hasattr(mod, classname):
                        cls_ = getattr(mod, classname)
                        data = {k: v for k, v in d.items() if not k.startswith("@")}
                        if hasattr(cls_, "from_dict"):
                            return cls_.from_dict(data)
                        if issubclass(cls_, Enum):
                            return cls_(d["value"])
                        if pydantic is not None and issubclass(
                            cls_, pydantic.BaseModel
                        ):  # pylint: disable=E1101
                            d = {k: self.process_decoded(v) for k, v in data.items()}
                            return cls_(**d)
                        if (
                            dataclasses is not None
                            and (not issubclass(cls_, MSONable))
                            and dataclasses.is_dataclass(cls_)
                        ):
                            d = {k: self.process_decoded(v) for k, v in data.items()}
                            return cls_(**d)

                elif torch is not None and modname == "torch" and classname == "Tensor":
                    if "Complex" in d["dtype"]:
                        return torch.tensor(  # pylint: disable=E1101
                            [
                                np.array(r) + np.array(i) * 1j
                                for r, i in zip(*d["data"])
                            ],
                        ).type(d["dtype"])
                    return torch.tensor(d["data"]).type(d["dtype"])  # pylint: disable=E1101

                elif np is not None and modname == "numpy" and classname == "array":
                    if d["dtype"].startswith("complex"):
                        return np.array(
                            [
                                np.array(r) + np.array(i) * 1j
                                for r, i in zip(*d["data"])
                            ],
                            dtype=d["dtype"],
                        )
                    return np.array(d["data"], dtype=d["dtype"])

                elif modname == "pandas":
                    import pandas as pd

                    if classname == "DataFrame":
                        decoded_data = MontyDecoder().decode(d["data"])
                        return pd.DataFrame(decoded_data)
                    if classname == "Series":
                        decoded_data = MontyDecoder().decode(d["data"])
                        return pd.Series(decoded_data)

                elif (
                    (bson is not None)
                    and modname == "bson.objectid"
                    and classname == "ObjectId"
                ):
                    return bson.objectid.ObjectId(d["oid"])

            return {
                self.process_decoded(k): self.process_decoded(v) for k, v in d.items()
            }

        if isinstance(d, list):
            return [self.process_decoded(x) for x in d]

        return d

    def decode(self, s):
        """
        Overrides decode from JSONDecoder.

        :param s: string
        :return: Object.
        """
        if orjson is not None:
            try:
                d = orjson.loads(s)  # pylint: disable=E1101
            except orjson.JSONDecodeError:  # pylint: disable=E1101
                d = json.loads(s)
        else:
            d = json.loads(s)
        return self.process_decoded(d)


class MSONError(Exception):
    """
    Exception class for serialization errors.
    """


def jsanitize(
    obj,
    strict=False,
    allow_bson=False,
    enum_values=False,
    recursive_msonable=False,
):
    """
    This method cleans an input json-like object, either a list or a dict or
    some sequence, nested or otherwise, by converting all non-string
    dictionary keys (such as int and float) to strings, and also recursively
    encodes all objects using Monty's as_dict() protocol.

    Args:
        obj: input json-like object.
        strict (bool): This parameter sets the behavior when jsanitize
            encounters an object it does not understand. If strict is True,
            jsanitize will try to get the as_dict() attribute of the object. If
            no such attribute is found, an attribute error will be thrown. If
            strict is False, jsanitize will simply call str(object) to convert
            the object to a string representation.
        allow_bson (bool): This parameter sets the behavior when jsanitize
            encounters a bson supported type such as objectid and datetime. If
            True, such bson types will be ignored, allowing for proper
            insertion into MongoDB databases.
        enum_values (bool): Convert Enums to their values.
        recursive_msonable (bool): If True, uses .as_dict() for MSONables regardless
            of the value of strict.

    Returns:
        Sanitized dict that can be json serialized.
    """
    if isinstance(obj, Enum):
        if enum_values:
            return obj.value
        elif hasattr(obj, "as_dict"):
            return obj.as_dict()
        return MontyEncoder().default(obj)

    if allow_bson and (
        isinstance(obj, (datetime.datetime, bytes))
        or (bson is not None and isinstance(obj, bson.objectid.ObjectId))
    ):
        return obj
    if isinstance(obj, (list, tuple)):
        return [
            jsanitize(
                i,
                strict=strict,
                allow_bson=allow_bson,
                enum_values=enum_values,
                recursive_msonable=recursive_msonable,
            )
            for i in obj
        ]
    if np is not None and isinstance(obj, np.ndarray):
        try:
            return [
                jsanitize(
                    i,
                    strict=strict,
                    allow_bson=allow_bson,
                    enum_values=enum_values,
                    recursive_msonable=recursive_msonable,
                )
                for i in obj.tolist()
            ]
        except TypeError:
            return obj.tolist()
    if np is not None and isinstance(obj, np.generic):
        return obj.item()
    if _check_type(
        obj,
        (
            "pandas.core.series.Series",
            "pandas.core.frame.DataFrame",
            "pandas.core.base.PandasObject",
        ),
    ):
        return obj.to_dict()
    if isinstance(obj, dict):
        return {
            str(k): jsanitize(
                v,
                strict=strict,
                allow_bson=allow_bson,
                enum_values=enum_values,
                recursive_msonable=recursive_msonable,
            )
            for k, v in obj.items()
        }
    if isinstance(obj, (int, float)):
        return obj
    if obj is None:
        return None
    if isinstance(obj, (pathlib.Path, datetime.datetime)):
        return str(obj)

    if callable(obj) and not isinstance(obj, MSONable):
        try:
            return _serialize_callable(obj)
        except TypeError:
            pass

    if recursive_msonable:
        try:
            return obj.as_dict()
        except AttributeError:
            pass

    if not strict:
        return str(obj)

    if isinstance(obj, str):
        return obj

    if pydantic is not None and isinstance(obj, pydantic.BaseModel):  # pylint: disable=E1101
        return jsanitize(
            MontyEncoder().default(obj),
            strict=strict,
            allow_bson=allow_bson,
            enum_values=enum_values,
            recursive_msonable=recursive_msonable,
        )

    return jsanitize(
        obj.as_dict(),
        strict=strict,
        allow_bson=allow_bson,
        enum_values=enum_values,
        recursive_msonable=recursive_msonable,
    )


def _serialize_callable(o):
    if isinstance(o, types.BuiltinFunctionType):
        # don't care about what builtin functions (sum, open, etc) are bound to
        bound = None
    else:
        # bound methods (i.e., instance methods) have a __self__ attribute
        # that points to the class/module/instance
        bound = getattr(o, "__self__", None)

    # we are only able to serialize bound methods if the object the method is
    # bound to is itself serializable
    if bound is not None:
        try:
            bound = MontyEncoder().default(bound)
        except TypeError:
            raise TypeError(
                "Only bound methods of classes or MSONable instances are supported."
            )

    return {
        "@module": o.__module__,
        "@callable": getattr(o, "__qualname__", o.__name__),
        "@bound": bound,
    }



"""
Useful additional functions to help get information about live objects
"""

from __future__ import annotations

import inspect
import os
from inspect import currentframe, getframeinfo
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Literal, Type


def all_subclasses(cls: Type) -> list[Type]:
    """
    Given a class `cls`, this recursive function returns a list with
    all subclasses, subclasses of subclasses, and so on.
    """
    subclasses = cls.__subclasses__()
    return subclasses + [g for s in subclasses for g in all_subclasses(s)]


def find_top_pyfile():
    """
    This function inspects the Cpython frame to find the path of the script.
    """
    frame = currentframe()
    while True:
        if frame.f_back is None:
            finfo = getframeinfo(frame)
            return os.path.abspath(finfo.filename)

        frame = frame.f_back


def caller_name(skip: Literal[1, 2] = 2) -> str:
    """
    Get a name of a caller in the format module.class.method

    `skip` specifies how many levels of stack to skip while getting caller
    name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.

    An empty string is returned if skipped levels exceed stack height

    Taken from:

        https://gist.github.com/techtonik/2151727

    Public Domain, i.e. feel free to copy/paste
    """
    stack = inspect.stack()
    start = 0 + skip
    if len(stack) < start + 1:
        return ""
    parentframe = stack[start][0]

    name = []

    # `modname` can be None when frame is executed directly in console
    # TODO(techtonik): consider using __main__
    if module := inspect.getmodule(parentframe):
        name.append(module.__name__)

    # detect classname
    if "self" in parentframe.f_locals:
        # I don't know any way to detect call from the object method
        # XXX: there seems to be no way to detect static method call - it will
        #      be just a function call
        name.append(parentframe.f_locals["self"].__class__.__name__)

    codename = parentframe.f_code.co_name
    if codename != "<module>":  # top level usually
        name.append(codename)  # function or a method
    del parentframe

    return ".".join(name)



"""
functools, especially backported from Python 3.
"""

from __future__ import annotations

import cProfile
import pstats
import signal
import sys
import tempfile
from collections import namedtuple
from functools import partial, wraps
from typing import Any, Callable, Union

_CacheInfo = namedtuple("_CacheInfo", ["hits", "misses", "maxsize", "currsize"])


class _HashedSeq(list):  # pylint: disable=C0205
    """
    This class guarantees that hash() will be called no more than once
    per element.  This is important because the lru_cache() will hash
    the key multiple times on a cache miss.
    """

    __slots__ = "hashvalue"

    def __init__(self, tup: tuple, hashfunc: Callable = hash) -> None:
        """
        Args:
            tup: Tuple.
            hashfunc: Hash function.
        """
        self[:] = tup
        self.hashvalue: int = hashfunc(tup)

    def __hash__(self) -> int:  # type: ignore[override]
        return self.hashvalue


def _make_key(
    args: tuple,
    kwds: dict,
    typed: bool,
    kwd_mark: tuple[object] = (object(),),
    fasttypes=None,
) -> Any:
    """
    Make a cache key from optionally typed positional and keyword arguments

    The key is constructed in a way that is flat as possible rather than
    as a nested structure that would take more memory.

    If there is only a single argument and its data type is known to cache
    its hash value, then that argument is returned without a wrapper.  This
    saves space and improves lookup speed.
    """
    key = args
    if kwds:
        sorted_items = sorted(kwds.items())
        key += kwd_mark
        for item in sorted_items:
            key += item
    if typed:
        key += tuple(type(v) for v in args)
        if kwds:
            key += tuple(type(v) for k, v in sorted_items)
    elif len(key) == 1 and isinstance(key[0], fasttypes):
        return key[0]
    return _HashedSeq(key)


class lazy_property:
    """
    lazy_property descriptor

    Used as a decorator to create lazy attributes. Lazy attributes
    are evaluated on first use.
    """

    def __init__(self, func: Callable) -> None:
        """
        Args:
            func: Function to decorate.
        """
        self.__func = func
        wraps(self.__func)(self)  # type: ignore

    def __get__(self, inst: Any, inst_cls) -> Any:
        if inst is None:
            return self

        if not hasattr(inst, "__dict__"):
            raise AttributeError(
                f"'{inst_cls.__name__}' object has no attribute '__dict__'"
            )

        name = self.__name__  # type: ignore  # pylint: disable=E1101
        if name.startswith("__") and not name.endswith("__"):
            name = f"_{inst_cls.__name__}{name}"

        value = self.__func(inst)
        inst.__dict__[name] = value
        return value

    @classmethod
    def invalidate(cls, inst: object, name: str) -> None:
        """Invalidate a lazy attribute.

        This obviously violates the lazy contract. A subclass of lazy
        may however have a contract where invalidation is appropriate.
        """
        inst_cls = inst.__class__

        if not hasattr(inst, "__dict__"):
            raise AttributeError(
                f"'{inst_cls.__name__}' object has no attribute '__dict__'"
            )

        if name.startswith("__") and not name.endswith("__"):
            name = f"_{inst_cls.__name__}{name}"

        if not isinstance(getattr(inst_cls, name), cls):
            raise AttributeError(
                f"'{inst_cls.__name__}.{name}' is not a {cls.__name__} attribute"
            )

        if name in inst.__dict__:
            del inst.__dict__[name]


def return_if_raise(
    exception_tuple: Union[list, tuple], retval_if_exc: Any, disabled: bool = False
) -> Any:
    """
    Decorator for functions, methods or properties. Execute the callable in a
    try block, and return retval_if_exc if one of the exceptions listed in
    exception_tuple is raised (se also ``return_node_if_raise``).

    Setting disabled to True disables the try except block (useful for
    debugging purposes). One can use this decorator to define properties.

    Examples:
        @return_if_raise(ValueError, None)
        def return_none_if_value_error(self):
            pass

        @return_if_raise((ValueError, KeyError), "hello")
        def another_method(self):
            pass

        @property
        @return_if_raise(AttributeError, None)
        def name(self):
            "Name of the object, None if not set."
            return self._name

    """
    # we need a tuple of exceptions.
    if isinstance(exception_tuple, list):
        exception_tuple = tuple(exception_tuple)
    elif not isinstance(exception_tuple, tuple):
        exception_tuple = (exception_tuple,)
    else:
        raise TypeError(f"Wrong exception_tuple {type(exception_tuple)}")

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if disabled:
                return func(*args, **kwargs)
            try:
                return func(*args, **kwargs)
            except exception_tuple:  # pylint: disable=E0712
                return retval_if_exc

        return wrapper

    return decorator


# One could use None as default value in return_if_raise but this one is
# explicit and  more readable
return_none_if_raise = partial(return_if_raise, retval_if_exc=None)
"""
This decorator returns None if one of the exceptions is raised.

    @return_none_if_raise(ValueError)
    def method(self):
"""


class timeout:
    """
    Timeout function. Use to limit matching to a certain time limit. Note that
    this works only on Unix-based systems as it uses signal.

    Usage:
        try:
            with timeout(3):
                do_stuff()
        except TimeoutError:
            do_something_else()
    """

    def __init__(self, seconds: int = 1, error_message: str = "Timeout"):
        """
        Args:
            seconds (int): Allowed time for function in seconds.
            error_message (str): An error message.

        """
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, signum, frame):
        """
        Args:
            signum: Return signal from call.
            frame:
        """
        raise TimeoutError(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, type, value, traceback):
        signal.alarm(0)


class TimeoutError(Exception):
    """
    Exception class for timeouts.
    """

    def __init__(self, message: str):
        """
        Args:
            message: Error message
        """
        self.message = message


def prof_main(main):
    """
    Decorator for profiling main programs.

    Profiling is activated by prepending the command line options
    supported by the original main program with the keyword `prof`.

    Examples:

            $ script.py arg --foo=1

        becomes

            $ script.py prof arg --foo=1

        The decorated main accepts two new arguments:

            prof_file: Name of the output file with profiling data
                If not given, a temporary file is created.
            sortby: Profiling data are sorted according to this value.
                default is "time". See sort_stats.
    """

    @wraps(main)
    def wrapper(*args, **kwargs):
        try:
            do_prof = sys.argv[1] == "prof"
            if do_prof:
                sys.argv.pop(1)
        except Exception:
            do_prof = False

        if not do_prof:
            sys.exit(main())
        else:
            print("Entering profiling mode...")
            prof_file = kwargs.get("prof_file", None)
            if prof_file is None:
                _, prof_file = tempfile.mkstemp()
                print(f"Profiling data stored in {prof_file}")

            sortby = kwargs.get("sortby", "time")
            cProfile.runctx("main()", globals(), locals(), prof_file)
            s = pstats.Stats(prof_file)
            s.strip_dirs().sort_stats(sortby).print_stats()
            if "retval" not in kwargs:
                sys.exit(0)
            else:
                return kwargs["retval"]

    return wrapper



"""
Some common design patterns such as singleton and cached classes.
"""

from __future__ import annotations

import inspect
import os
from functools import wraps
from typing import Any, Dict, Hashable, Tuple, TypeVar
from weakref import WeakValueDictionary


def singleton(cls):
    """
    This decorator can be used to create a singleton out of a class.

    Usage:
        @singleton
        class MySingleton():

            def __init__():
                pass
    """

    instances = {}

    def getinstance():
        if cls not in instances:
            instances[cls] = cls()
        return instances[cls]

    return getinstance


# https://github.com/microsoft/pylance-release/issues/3478
Klass = TypeVar("Klass")


def cached_class(cls: type[Klass]) -> type[Klass]:
    """
    Decorator to cache class instances by constructor arguments.
    This results in a class that behaves like a singleton for each
    set of constructor arguments, ensuring efficiency.

    Note that this should be used for *immutable classes only*.  Having
    a cached mutable class makes very little sense.  For efficiency,
    avoid using this decorator for situations where there are many
    constructor arguments permutations.

    If any arguments are non-hashable, that set of arguments
    is not cached.
    """
    orig_new = cls.__new__
    orig_init = cls.__init__
    cache: WeakValueDictionary = WeakValueDictionary()

    @wraps(orig_new)
    def new_new(cls, *args: Any, **kwargs: Any) -> Any:
        # Normalize arguments
        sig = inspect.signature(orig_init)
        bound_args = sig.bind(None, *args, **kwargs)
        bound_args.apply_defaults()

        # Remove 'self' from the arguments
        normalized_args = tuple(bound_args.arguments.values())[1:]

        try:
            key = (cls, normalized_args)
            if key in cache:
                return cache[key]

            if orig_new is object.__new__:
                instance = orig_new(cls)
            else:
                instance = orig_new(cls, *args, **kwargs)

            orig_init(instance, *args, **kwargs)
            instance._initialized = True
            cache[key] = instance
            return instance
        except TypeError:
            # Can't cache this set of arguments
            if orig_new is object.__new__:
                instance = orig_new(cls)
            else:
                instance = orig_new(cls, *args, **kwargs)
            orig_init(instance, *args, **kwargs)
            instance._initialized = True
            return instance

    @wraps(orig_init)
    def new_init(self: Any, *args: Any, **kwargs: Any) -> None:
        if not hasattr(self, "_initialized"):
            orig_init(self, *args, **kwargs)
            self._initialized = True

    def reduce(self: Any) -> Tuple[type, Tuple, Dict[str, Any]]:
        for key, value in cache.items():
            if value is self:
                cls, args = key
                return (cls, args, {})
        raise ValueError("Instance not found in cache")

    cls.__new__ = new_new  # type: ignore[method-assign]
    cls.__init__ = new_init  # type: ignore[method-assign]
    cls.__reduce__ = reduce  # type: ignore[method-assign]

    return cls


class NullFile:
    """A file object that is associated to /dev/null."""

    def __new__(cls):
        """
        Pass through.
        """
        return open(os.devnull, "w")  # pylint: disable=R1732

    def __init__(self):
        """no-op"""


class NullStream:
    """A fake stream with a no-op write."""

    def write(self, *args):  # pylint: disable=E0211
        """
        Does nothing...
        """



"""Copying and zipping utilities. Works on directories mostly."""

from __future__ import annotations

import os
import shutil
import warnings
from gzip import GzipFile
from pathlib import Path
from typing import TYPE_CHECKING

from monty.io import zopen

if TYPE_CHECKING:
    from typing import Literal, Optional


def copy_r(src: str | Path, dst: str | Path) -> None:
    """
    Implements a recursive copy function similar to Unix's "cp -r" command.
    Surprisingly, python does not have a real equivalent. shutil.copytree
    only works if the destination directory is not present.

    Args:
        src (str | Path): Source folder to copy.
        dst (str | Path): Destination folder.
    """
    src = Path(src)
    dst = Path(dst)
    abssrc = src.resolve()
    absdst = dst.resolve()
    os.makedirs(absdst, exist_ok=True)
    for filepath in os.listdir(abssrc):
        fpath = Path(abssrc, filepath)
        if fpath.is_symlink():
            continue
        if fpath.is_file():
            shutil.copy(fpath, absdst)
        elif str(fpath) not in str(absdst):
            copy_r(fpath, Path(absdst, filepath))
        else:
            warnings.warn(f"Cannot copy {fpath} to itself")


def gzip_dir(path: str | Path, compresslevel: int = 6) -> None:
    """
    Gzips all files in a directory. Note that this is different from
    shutil.make_archive, which creates a tar archive. The aim of this method
    is to create gzipped files that can still be read using common Unix-style
    commands like zless or zcat.

    Args:
        path (str | Path): Path to directory.
        compresslevel (int): Level of compression, 1-9. 9 is default for
            GzipFile, 6 is default for gzip.
    """
    path = Path(path)
    for root, _, files in os.walk(path):
        for f in files:
            full_f = Path(root, f).resolve()
            if Path(f).suffix.lower() != ".gz" and not full_f.is_dir():
                with (
                    open(full_f, "rb") as f_in,
                    GzipFile(
                        f"{full_f}.gz", "wb", compresslevel=compresslevel
                    ) as f_out,
                ):
                    shutil.copyfileobj(f_in, f_out)
                shutil.copystat(full_f, f"{full_f}.gz")
                os.remove(full_f)


def compress_file(
    filepath: str | Path,
    compression: Literal["gz", "bz2"] = "gz",
    target_dir: Optional[str | Path] = None,
) -> None:
    """
    Compresses a file with the correct extension. Functions like standard
    Unix command line gzip and bzip2 in the sense that the original
    uncompressed files are not retained.

    Args:
        filepath (str | Path): Path to file.
        compression (str): A compression mode. Valid options are "gz" or
            "bz2". Defaults to "gz".
        target_dir (str | Path): An optional target dir where the result compressed
            file would be stored. Defaults to None for in-place compression.
    """
    filepath = Path(filepath)
    target_dir = Path(target_dir) if target_dir is not None else None

    if compression not in {"gz", "bz2"}:
        raise ValueError("Supported compression formats are 'gz' and 'bz2'.")

    if filepath.suffix.lower() != f".{compression}" and not filepath.is_symlink():
        if target_dir is not None:
            os.makedirs(target_dir, exist_ok=True)
            compressed_file: str | Path = target_dir / f"{filepath.name}.{compression}"

        else:
            compressed_file = f"{str(filepath)}.{compression}"

        with open(filepath, "rb") as f_in, zopen(compressed_file, "wb") as f_out:
            f_out.writelines(f_in)

        os.remove(filepath)


def compress_dir(path: str | Path, compression: Literal["gz", "bz2"] = "gz") -> None:
    """
    Recursively compresses all files in a directory. Note that this
    compresses all files singly, i.e., it does not create a tar archive. For
    that, just use Python tarfile class.

    Args:
        path (str | Path): Path to parent directory.
        compression (str): A compression mode. Valid options are "gz" or
            "bz2". Defaults to gz.
    """
    path = Path(path)
    for parent, _, files in os.walk(path):
        for f in files:
            compress_file(Path(parent, f), compression=compression)

    return


def decompress_file(
    filepath: str | Path, target_dir: Optional[str | Path] = None
) -> str | None:
    """
    Decompresses a file with the correct extension. Automatically detects
    gz, bz2 or z extension.

    Args:
        filepath (str | Path): Path to file.
        target_dir (str | Path): An optional target dir where the result decompressed
            file would be stored. Defaults to None for in-place decompression.

    Returns:
        str | None: The decompressed file path, None if no operation.
    """
    filepath = Path(filepath)
    target_dir = Path(target_dir) if target_dir is not None else None
    file_ext = filepath.suffix

    if file_ext.lower() in {".bz2", ".gz", ".z"} and filepath.is_file():
        if target_dir is not None:
            os.makedirs(target_dir, exist_ok=True)
            decompressed_file: str | Path = target_dir / filepath.name.removesuffix(
                file_ext
            )
        else:
            decompressed_file = str(filepath).removesuffix(file_ext)

        with zopen(filepath, "rb") as f_in, open(decompressed_file, "wb") as f_out:
            f_out.writelines(f_in)

        os.remove(filepath)

        return str(decompressed_file)
    return None


def decompress_dir(path: str | Path) -> None:
    """
    Recursively decompresses all files in a directory.

    Args:
        path (str | Path): Path to parent directory.
    """
    path = Path(path)
    for parent, _, files in os.walk(path):
        for f in files:
            decompress_file(Path(parent, f))


def remove(path: str | Path, follow_symlink: bool = False) -> None:
    """
    Implements a remove function that will delete files, folder trees and
    symlink trees.

    1.) Remove a file
    2.) Remove a symlink and follow into with a recursive rm if follow_symlink
    3.) Remove directory with rmtree

    Args:
        path (str | Path): path to remove
        follow_symlink(bool): follow symlinks and removes whatever is in them
    """
    path = Path(path)
    if path.is_file():
        os.remove(path)
    elif path.is_symlink():
        if follow_symlink:
            remove(os.readlink(path))
        Path.unlink(path)
    else:
        shutil.rmtree(path)



"""
Helpful regex based functions. E.g., grepping.
"""

from __future__ import annotations

import collections
import contextlib
import re
from typing import TYPE_CHECKING

from monty.io import reverse_readfile, zopen

if TYPE_CHECKING:
    from typing import Callable


def regrep(
    filename: str,
    patterns: dict,
    reverse: bool = False,
    terminate_on_match: bool = False,
    postprocess: Callable = str,
) -> dict:
    r"""
    A powerful regular expression version of grep.

    Args:
        filename (str): Filename to grep.
        patterns (dict): A dict of patterns, e.g.,
            {"energy": r"energy\\(sigma->0\\)\\s+=\\s+([\\d\\-\\.]+)"}.
        reverse (bool): Read files in reverse. Defaults to false. Useful for
            large files, especially when used with terminate_on_match.
        terminate_on_match (bool): Whether to terminate when there is at
            least one match in each key in pattern.
        postprocess (callable): A post processing function to convert all
            matches. Defaults to str, i.e., no change.

    Returns:
        A dict of the following form:
            {key1: [[[matches...], lineno], [[matches...], lineno],
                    [[matches...], lineno], ...],
            key2: ...}
        For reverse reads, the lineno is given as a -ve number. Please note
        that 0-based indexing is used.
    """
    compiled = {k: re.compile(v) for k, v in patterns.items()}
    matches = collections.defaultdict(list)
    gen = reverse_readfile(filename) if reverse else zopen(filename, "rt")
    for i, line in enumerate(gen):
        for k, p in compiled.items():
            if m := p.search(line):
                matches[k].append(
                    [[postprocess(g) for g in m.groups()], -i if reverse else i]
                )
        if terminate_on_match and all(len(matches.get(k, [])) for k in compiled):
            break

    with contextlib.suppress(Exception):
        # Try to close open file handle. Pass if it is a generator.
        gen.close()  # type: ignore[attr-defined]
    return matches



"""
Calling shell processes.
"""

from __future__ import annotations

import shlex
import threading
import traceback
from subprocess import PIPE, Popen
from typing import TYPE_CHECKING

from monty.string import is_string

if TYPE_CHECKING:
    from typing import Optional

    from typing_extensions import Self

__author__ = "Matteo Giantomass"
__copyright__ = "Copyright 2014, The Materials Virtual Lab"
__version__ = "0.1"
__maintainer__ = "Matteo Giantomassi"
__email__ = "gmatteo@gmail.com"
__date__ = "10/26/14"


class Command:
    """
    Enables to run subprocess commands in a different thread with TIMEOUT
    option.

    Based on jcollado's solution:
        http://stackoverflow.com/questions/1191374/subprocess-with-timeout/4825933#4825933
    and
        https://gist.github.com/kirpit/1306188


    Attributes:
        retcode: Return code of the subprocess
        killed: True if subprocess has been killed due to the timeout
        output: stdout of the subprocess
        error: stderr of the subprocess

    Examples:
        com = Command("sleep 1").run(timeout=2)
        print(com.retcode, com.killed, com.output, com.output)
    """

    def __init__(self, command: str):
        """
        Args:
            command: Command to execute
        """
        if is_string(command):
            _command: list[str] = shlex.split(command)
        self.command = _command
        self.process = None
        self.retcode = None
        self.output, self.error = "", ""
        self.killed = False

    def __str__(self):
        return f"command: {self.command}, retcode: {self.retcode}"

    def run(self, timeout: Optional[float] = None, **kwargs) -> Self:
        """
        Run a command in a separated thread and wait timeout seconds.
        kwargs are keyword arguments passed to Popen.

        Returns:
            self
        """

        def target(**kw):
            try:
                # print('Thread started')
                with Popen(self.command, **kw) as self.process:
                    self.output, self.error = self.process.communicate()
                    self.retcode = self.process.returncode
                # print('Thread stopped')
            except Exception:
                self.error = traceback.format_exc()
                self.retcode = -1

        # default stdout and stderr
        if "stdout" not in kwargs:
            kwargs["stdout"] = PIPE

        if "stderr" not in kwargs:
            kwargs["stderr"] = PIPE

        # thread
        thread = threading.Thread(target=target, kwargs=kwargs)
        thread.start()
        thread.join(timeout)

        if thread.is_alive() and self.process is not None:
            # print("Terminating process")
            self.process.terminate()
            self.killed = True
            thread.join()

        return self



"""
Math functions.
"""

from __future__ import annotations

import math
from typing import Sequence


def gcd(*numbers: int) -> int:
    """
    Returns the greatest common divisor for a sequence of numbers.

    Args:
        *numbers: Sequence of numbers.

    Returns:
        int: Greatest common divisor of numbers.
    """
    n: int = numbers[0]
    for i in numbers:
        n = math.gcd(n, i)
    return n


def lcm(*numbers: int) -> int:
    """
    Return lowest common multiple of a sequence of numbers.

    Args:
        *numbers: Sequence of numbers.

    Returns:
        int: Lowest common multiple of numbers.
    """
    n = 1
    for i in numbers:
        n = (i * n) // gcd(i, n)
    return n


def gcd_float(numbers: Sequence[float], tol: float = 1e-8) -> float:
    """
    Returns the greatest common divisor for a sequence of numbers.
    Uses a numerical tolerance, so can be used on floats

    Args:
        numbers: Sequence of numbers.
        tol: Numerical tolerance

    Returns:
        float: Greatest common divisor of numbers.
    """

    def pair_gcd_tol(a: float, b: float) -> float:
        """Calculate the Greatest Common Divisor of a and b.

        Unless b==0, the result will have the same sign as b (so that when
        b is divided by it, the result comes out positive).
        """
        while b > tol:
            a, b = b, a % b
        return a

    n = numbers[0]
    for i in numbers:
        n = pair_gcd_tol(n, i)
    return n



"""Deployment file to facilitate releases of monty."""

from __future__ import annotations

import datetime
import glob
import json
import os
import re
from typing import TYPE_CHECKING

import requests
from invoke import task

from monty import __version__ as ver
from monty.os import cd

if TYPE_CHECKING:
    from invoke import Context

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyue@mit.edu"
__date__ = "Apr 29, 2012"

NEW_VER = datetime.datetime.today().strftime("%Y.%-m.%-d")


@task
def make_doc(ctx: Context) -> None:
    with cd("docs"):
        ctx.run("rm monty.*.rst", warn=True)
        ctx.run("sphinx-apidoc --separate -P -M -d 6 -o . -f ../monty")
        # ctx.run("rm monty*.html", warn=True)
        # ctx.run("sphinx-build -b html . ../docs")  # HTML building.
        ctx.run("sphinx-build -M markdown . .")
        ctx.run("rm *.rst", warn=True)
        ctx.run("cp markdown/monty*.md .")
        for fn in glob.glob("monty*.md"):
            with open(fn) as f:
                lines = [line.rstrip() for line in f if "Submodules" not in line]
            if fn == "monty.md":
                preamble = [
                    "---",
                    "layout: default",
                    "title: API Documentation",
                    "nav_order: 5",
                    "---",
                    "",
                ]
            else:
                preamble = [
                    "---",
                    "layout: default",
                    f"title: {fn}",
                    "nav_exclude: true",
                    "---",
                    "",
                ]
            with open(fn, "w") as f:
                f.write("\n".join(preamble + lines))

        ctx.run("rm -r markdown", warn=True)
        ctx.run("cp ../*.md .")
        ctx.run("mv README.md index.md")
        ctx.run("rm -rf *.orig doctrees", warn=True)

        with open("index.md") as f:
            contents = f.read()
        with open("index.md", "w") as f:
            contents = re.sub(
                r"\n## Official Documentation[^#]*",
                "{: .no_toc }\n\n## Table of contents\n{: .no_toc .text-delta }\n* TOC\n{:toc}\n\n",
                contents,
            )
            contents = (
                "---\nlayout: default\ntitle: Home\nnav_order: 1\n---\n\n" + contents
            )

            f.write(contents)


@task
def update_doc(ctx: Context) -> None:
    ctx.run("git pull", warn=True)
    make_doc(ctx)
    ctx.run("git add .", warn=True)
    ctx.run('git commit -a -m "Update dev docs"', warn=True)
    ctx.run("git push", warn=True)


@task
def test(ctx: Context) -> None:
    ctx.run("pytest")


@task
def setver(ctx: Context) -> None:
    ctx.run(f'sed s/version=.*,/version=\\"{ver}\\",/ setup.py > newsetup')
    ctx.run("mv newsetup setup.py")


@task
def release_github(ctx: Context) -> None:
    with open("docs/changelog.md") as f:
        contents = f.read()
    toks = re.split("##", contents)
    desc = toks[1].strip()
    payload = {
        "tag_name": f"v{NEW_VER}",
        "target_commitish": "master",
        "name": f"v{NEW_VER}",
        "body": desc,
        "draft": False,
        "prerelease": False,
    }

    response = requests.post(
        "https://api.github.com/repos/materialsvirtuallab/monty/releases",
        data=json.dumps(payload),
        headers={"Authorization": "token " + os.environ["GITHUB_RELEASES_TOKEN"]},
    )
    print(response.text)


@task
def commit(ctx: Context) -> None:
    ctx.run(f'git commit -a -m "v{NEW_VER} release"', warn=True)
    ctx.run("git push", warn=True)


@task
def set_ver(ctx: Context) -> None:
    with open("monty/__init__.py", encoding="utf-8") as f:
        contents = f.read()
        contents = re.sub(
            r"__version__ = .*\n", '__version__ = "%s"\n' % NEW_VER, contents
        )

    with open("monty/__init__.py", "w", encoding="utf-8") as f:
        f.write(contents)

    with open("pyproject.toml", encoding="utf-8") as f:
        contents = f.read()
        contents = re.sub(r"version = ([\.\d\"]+)", f'version = "{NEW_VER}"', contents)

    with open("pyproject.toml", "w", encoding="utf-8") as f:
        f.write(contents)


@task
def release(ctx: Context, notest: bool = False) -> None:
    set_ver(ctx)
    if not notest:
        test(ctx)
    update_doc(ctx)
    commit(ctx)
    release_github(ctx)
    ctx.run("python -m build", warn=True)
    ctx.run("python -m build --wheel", warn=True)
    ctx.run("twine upload --skip-existing dist/*.whl", warn=True)
    ctx.run("twine upload --skip-existing dist/*.tar.gz", warn=True)



"""
This module implements several useful functions and decorators that can be
particularly useful for developers. E.g., deprecating methods / classes, etc.
"""

from __future__ import annotations

import functools
import inspect
import logging
import os
import subprocess
import sys
import warnings
from datetime import datetime
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Callable, Optional, Type

logger = logging.getLogger(__name__)


def deprecated(
    replacement: Optional[Callable | str] = None,
    message: str = "",
    deadline: Optional[tuple[int, int, int]] = None,
    category: Type[Warning] = FutureWarning,
) -> Callable:
    """
    Decorator to mark classes or functions as deprecated, with a possible replacement.

    Args:
        replacement (Callable | str): A replacement class or function.
        message (str): A warning message to be displayed.
        deadline (Optional[tuple[int, int, int]]): Optional deadline for removal
            of the old function/class, in format (yyyy, MM, dd). A CI warning would
            be raised after this date if is running in code owner' repo.
        category (Warning): Choose the category of the warning to issue. Defaults
            to FutureWarning. Another choice can be DeprecationWarning. Note that
            FutureWarning is meant for end users and is always shown unless silenced.
            DeprecationWarning is meant for developers and is never shown unless
            python is run in developmental mode or the filter is changed. Make
            the choice accordingly.

    Returns:
        Original function, but with a warning to use the updated function.
    """

    def raise_deadline_warning() -> None:
        """Raise CI warning after removal deadline in code owner's repo."""

        def _is_in_owner_repo() -> bool:
            """Check if is running in code owner's repo.
            Only generate reliable check when `git` is installed and remote name
            is "origin".
            """

            try:
                # Get current running repo
                result = subprocess.run(
                    ["git", "config", "--get", "remote.origin.url"],
                    stdout=subprocess.PIPE,
                )
                owner_repo = (
                    result.stdout.decode("utf-8")
                    .strip()
                    .lstrip("https://github.com/")  # HTTPS clone
                    .lstrip("git@github.com:")  # SSH clone
                    .rstrip(".git")  # SSH clone
                )

                return owner_repo == os.getenv("GITHUB_REPOSITORY")

            except (subprocess.CalledProcessError, FileNotFoundError):
                return False

        # Only raise warning in code owner's repo CI
        if (
            _deadline is not None
            and os.getenv("CI") is not None
            and datetime.now() > _deadline
            and _is_in_owner_repo()
        ):
            raise DeprecationWarning(
                f"This function should have been removed on {_deadline:%Y-%m-%d}."
            )

    def craft_message(
        old: Callable,
        replacement: Callable | str,
        message: str,
        deadline: datetime,
    ) -> str:
        msg = f"{old.__name__} is deprecated"

        if deadline is not None:
            msg += f", and will be removed on {_deadline:%Y-%m-%d}\n"

        if replacement is not None:
            if deadline is None:
                msg += "; use "  # for better formatting
            else:
                msg += "Use "

            if isinstance(replacement, str):
                msg += f"{replacement} instead."

            else:
                if isinstance(replacement, property):
                    r = replacement.fget
                elif isinstance(replacement, (classmethod, staticmethod)):
                    r = replacement.__func__
                else:
                    r = replacement

                msg += f"{r.__name__} in {r.__module__} instead."

        if message:
            msg += "\n" + message
        return msg

    def deprecated_function_decorator(old: Callable) -> Callable:
        def wrapped(*args, **kwargs):
            msg = craft_message(old, replacement, message, _deadline)
            warnings.warn(msg, category=category, stacklevel=2)
            return old(*args, **kwargs)

        return wrapped

    def deprecated_class_decorator(cls: Type) -> Type:
        original_init = cls.__init__

        def new_init(self, *args, **kwargs):
            msg = craft_message(cls, replacement, message, _deadline)
            warnings.warn(msg, category=category, stacklevel=2)
            original_init(self, *args, **kwargs)

        cls.__init__ = new_init
        return cls

    # Convert deadline to datetime type
    _deadline = datetime(*deadline) if deadline is not None else None

    # Raise CI warning after removal deadline
    raise_deadline_warning()

    def decorator(target: Callable) -> Callable:
        if inspect.isfunction(target):
            return deprecated_function_decorator(target)
        elif inspect.isclass(target):
            return deprecated_class_decorator(target)
        else:
            raise TypeError(
                "The @deprecated decorator can only be applied to classes or functions"
            )

    return decorator


class requires:
    """
    Decorator to mark classes or functions as requiring a specified condition
    to be true. This can be used to present useful error messages for
    optional dependencies. For example, decorating the following code will
    check if scipy is present and if not, a runtime error will be raised if
    someone attempts to call the use_scipy function:

        try:
            import scipy
        except ImportError:
            scipy = None

        @requires(scipy is not None, "scipy is not present.")
        def use_scipy():
            print(scipy.majver)

    Args:
        condition: Condition necessary to use the class or function.
        message: A message to be displayed if the condition is not True.
    """

    def __init__(
        self, condition: bool, message: str, err_cls: type[Exception] = RuntimeError
    ) -> None:
        """
        Args:
            condition: A expression returning a bool.
            message: Message to display if condition is False.
        """
        self.condition = condition
        self.message = message
        self.err_cls = err_cls

    def __call__(self, _callable: Callable) -> Callable:
        """
        Args:
            _callable: Callable function.
        """

        @functools.wraps(_callable)
        def decorated(*args, **kwargs):
            if not self.condition:
                raise self.err_cls(self.message)
            return _callable(*args, **kwargs)

        return decorated


def install_excepthook(hook_type: str = "color", **kwargs) -> int:
    """
    This function replaces the original python traceback with an improved
    version from Ipython. Use `color` for colourful traceback formatting,
    `verbose` for Ka-Ping Yee's "cgitb.py" version kwargs are the keyword
    arguments passed to the constructor. See IPython.core.ultratb.py for more
    info.

    Returns:
        0 if hook is installed successfully.
    """
    try:
        from IPython.core import ultratb  # pylint: disable=import-outside-toplevel
    except ImportError:
        warnings.warn("Cannot install excepthook, IPyhon.core.ultratb not available")
        return 1

    # Select the hook.
    hook = dict(
        color=ultratb.ColorTB,
        verbose=ultratb.VerboseTB,
    ).get(hook_type.lower(), None)

    if hook is None:
        return 2

    sys.excepthook = hook(**kwargs)
    return 0
